# Copyright (C) 2010 Alejandro Ochoa, Singh Research Group, Princeton University
# This file is part of dPUC.
# dPUC is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# dPUC is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with dPUC.  If not, see <http://www.gnu.org/licenses/>.

package Context;
use Inline C => './Context.c', 'NAME' => 'Context';
use IPC::Run ();
use lib '.';
use General;
use Hits;
use Nets;
use strict;

# Note on negative scores: We decided to implement these two rules to prevent funny things from happening:
#   a) If we have pair counts cij > 0 but the score is negative (Sij < 0, happens very infrequently), then we set Sij = 0 (ignore edge).
#   b) If we don't have pair counts cij = 0, but the score is positive (Sij > 0, happens with a much larger frequency), then we set Sij = 0 (ignore edge).
# Essentially we want things with counts to be negative, and things without counts to be negative, and the cases that don't satisfy that are just wrong.  We avoid them with the above rules.

# introduction of C code changes a few things further
# - context scores are internally integers, scaled versions of the original context scores so that we retain some precision (one or two decimal points).
#   This should be fine as long as scores remain low (current context score max is ~20), and you're using a decent computer/compiler (where ints use 32 bits).
#   Range of 32-bit signed ints: -2,147,483,648 to +2,147,483,647
#   Current code doesn't check for over/underflows, but they really shouldn't happen for biologically realistic proteins
#   (given the sizes of the individual context scores and the amount of domains that proteins usually have).

# package constants
my $scoreScale = 1000; # in posElim, we multiply context scores internally by this much, then turn into ints (so additions are faster), but we also have to keep this scale in mind when we combine context and HMM scores!  This loses some precision, but I think it's acceptable for the sake of speed, and also given the uncertainty involved in the actual calculation of context scores.
# lp_solve outcome constants used in sub readLpSol
our $optimalStr = 'Optimal';
our $suboptimalStr = 'Suboptimal';
our $timeoutStr = 'Timeout';
my $mallocLongStr = quotemeta '*** malloc'; # this one is for internal use only
our $mallocShortStr = 'MallocError';

sub runHmmerPipe {
    # this awesome piece of code runs HMMER and gets results without having to write temporary files for our FASTA file or the HMMER output
    
    # get relevant parameters for this query
    my ($id, $seq, $Ecut, $pfamLib, $hmmpfam) = @_;
    # make text input for this sequence (FASTA format, HMMER doesn't like getting the sequence alone)
    my $fa = ">$id\n$seq\n";
    # command and parameters (the last - means sequence will come from STDIN)
    my @hmmer = ($hmmpfam, '--acc', '-A', '0', '-E', $Ecut, $pfamLib, '-');
    # output, undefined until it returns
    my $out;
    # execute via IPC::Run, which allows us to pipe data in and out of hmmer!!!
    IPC::Run::run(\@hmmer, \$fa, '>&', \$out); # $fa is STDIN, $out will capture both STDOUT and STDERR!
    # done, return output.  We'll have to decide outside if there were errors or not...
    return $out;
}

sub loadContextNet {
    # this is a wrapper around the file that normally loads the net, taking all parameters into account, and also includes the typical pre-processing that is necessary for posElim or nonOvNeg
    # $isPosElim is a boolean, true means posElim, false means nonOvNeg
    my ($isPosElim, $fiNet) = @_;
    
    # now actually load net into structure, filter if necessary
    # we only want positive scores if doing posElim.  We need the complete list (cut = undef) if nonOvNeg
    my $scoreCut = $isPosElim ? 0 : undef;
    my $net = Nets::loadNet($fiNet, $scoreCut);
    
    # when doing posElim, scale scores to ints, turn into parallel arrays, store in C space (so fast code access it fast)
    if ($isPosElim) {
	# scale to have integers, trying to speed up calculations
	Nets::scaleNetAndInt($net, $scoreScale);
	  # this turns double hash net into a parallel array that is a lot more C-friendly, but not perlish at all
	  my ($netTs, $netTscores) = Nets::mapPfamNetAccs2ints($net);
	  # this copies these perl arrays into global and native C arrays, should speed a few things up
	  copyNetTscoresToC($netTs, $netTscores);
      }
    
    # both posElim and nonOvNeg need this step
    # getting all edges both ways makes it much faster to browse pairs in Perl
    my $netBothWays = Nets::makeAllEdgesBothWays($net);
    
    # posElim needs a list of domains that carry context information, also gets stored in C space
    if ($isPosElim) {
	# getting list of Pfam accessions that carry context information
	my @accsCc = keys %$netBothWays; # not sorted yet
	# decided to store as ints rather than whole strings, in order to do binary searches in C
	my @accIntsCc;
	foreach my $acc (@accsCc) {
	    my $accInt = $acc;
	    $accInt =~ s/^PF//; # should leave an integer behind only, works for Pfam only!!!
	    push @accIntsCc, $accInt;
	}
	@accIntsCc = sort { $a <=> $b } @accIntsCc; # now sort with perl, numerically ascending
	copyAccIntsCcToC(\@accIntsCc); # lastly, copy to global and native C array
    }
    
    # both posElim and nonOvNeg use this at some point, so return it
    return $netBothWays;
}

sub lpSolveNonOvNeg {
    # this is the generalized sub that runs lpSolve with nonOvNeg on a set of hits
    # problem is always passed in a pipe, and output captured from STDOUT, so entire process does not write/read any files.  Should be a lot faster on systems with slow I/O (like clusters?), and it also saves a lot of space since we don't write a lot of little files for each problem.  The only drawback is that solution has to be parsed and printed to its destination right away
    # as a consequence, unlike previous versions, this one always runs lp_solve, since there are no previous solution files to parse
    # all inputs are mandatory except for $acc2clan, it's only used when parsing incomplete runs and it's not that important anyway
    my ($hits, $netBothWays, $nestingNet, $acc2mode2ds2t, $acc2mode2seqBol, $dt2E, $Emax, $timeout) = @_;
    
    # somewhat often there are no input domains, bail out early in that case!
    return ($hits, 'NoHits') unless @$hits;
    
    # this handy sub transforms a single @$hits problem, given the general scoring structures, into the final lp_solve problem as text
    my ($lp, $scoresNet) = hits2lpSolveNonOvNegText($hits, $netBothWays, $nestingNet, $acc2mode2ds2t, $acc2mode2seqBol, $dt2E, $Emax); 
    # execute via IPC::Run, which allows us to pipe data in and out of lp_solve!!!
    my @lpSolve = ($FilePaths::lpSolve, '-time', '-timeout', $timeout); # program and arguments
    my $sol; # undefined until it returns
    IPC::Run::run(\@lpSolve, \$lp, '>&', \$sol); # $lp is STDIN, $sol will capture both STDOUT and STDERR!
    # now we have to parse the output
    my ($hitsPass, $outcome) = readLpSol($sol, $hits);
    
    # if we had an optimal or suboptimal solution, or something else that wasn't timeout (shouldn't happen unless we had problems parsing the file), return what we got after wrapping up
    if ($outcome ne $timeoutStr) {
	# we have arrived at our final list of hits, store final scores
	$hitsPass = wrapUp_nonOvNeg($hitsPass, $scoresNet); # should be edited in reference
	# done, return new hits and status if run
	return ($hitsPass, $outcome);
    }
    # this should only be the timeout case
    else {
	# let's try to reasonably simplify the problem without just returning GA...
	# let's remove all overlaps while keeping the top scoring hits, then run lp_solve again
	my $hitsNonOv = Hits::removeOverlapsByScore($hits, $nestingNet); # remove overlaps as desired, also sorts hits by start
	my ($lpNonOv, $scoresNetNonOv) = hits2lpSolveNonOvNegText($hitsNonOv, $netBothWays, $nestingNet, $acc2mode2ds2t, $acc2mode2seqBol, $dt2E, $Emax); # text of problem
	my $solNonOv; # undefined for now, will have text output
	IPC::Run::run(\@lpSolve, \$lpNonOv, '>&', \$solNonOv); # run simpler problem, note we use the same timeout as before!
	my ($hitsPassNonOv, $outcomeNonOv) = readLpSol($solNonOv, $hitsNonOv); # read solution to new, simpler problem
	# if this problem was solved at any capacity (optimally or suboptimally), return that
	if ($outcomeNonOv ne $timeoutStr) {
	    $hitsPassNonOv = wrapUp_nonOvNeg($hitsPassNonOv, $scoresNetNonOv); # wrap up new hits with new scoresNet
	    return ($hitsPassNonOv, $outcome.','.$outcomeNonOv); # return, but outcome is now the combination of the original and the second outcome, just to be informative
	}
	# if we still couldn't solve this problem (double timeout?), return top non-overlapping GA, even if it's empty
	# but we still want to process cause negative context will bump down some GA...
	else {
	    my $hitsNonOvGa = Hits::filterHitsBooleanCol($hitsNonOv, 'GA'); # hits were already sorted by start, and this operation does not change sorting, so this is it
	    my ($lpNonOvGa, $scoresNetNonOvGa) = hits2lpSolveNonOvNegText($hitsNonOvGa, $netBothWays, $nestingNet, $acc2mode2ds2t, $acc2mode2seqBol, $dt2E, $Emax); # text of problem
	    my $solNonOvGa; # undefined for now, will have text output
	    IPC::Run::run(\@lpSolve, \$lpNonOvGa, '>&', \$solNonOvGa); # run simpler problem, note we use the same timeout as before!
	    my ($hitsPassNonOvGa, $outcomeNonOvGa) = readLpSol($solNonOvGa, $hitsNonOvGa); # read solution to new, simpler problem
	    # return whatever we got
	    if ($outcomeNonOvGa ne $timeoutStr) {
		$hitsPassNonOvGa = wrapUp_nonOvNeg($hitsPassNonOvGa, $scoresNetNonOvGa); # wrap up new hits with new scoresNet
		return ($hitsPassNonOvGa, $outcome.','.$outcomeNonOv.','.$outcomeNonOvGa); # return, but outcome is now the combination of the original, second, and third outcomes, just to be informative
	    }
	    # I'll be damned, but if this doesn't get solved either, just return GA without negative context processing
	    else {
		$hitsNonOvGa = wrapUp_nonOvNeg($hitsNonOvGa, $scoresNetNonOv); # wrap up new hits with new scoresNet that goes with this set of hits, have to down-level for a bit
		return ($hitsNonOvGa, $outcome.','.$outcomeNonOv.','.$outcomeNonOvGa); # return, but outcome is now the combination of the original, second, and third outcomes, just to be informative
	    }
	}
    }
}

sub hits2lpSolveNonOvNegText {
    # this sub takes a particular hits structure, and a few general data structures, and returns the text corresponding to the lp_solve problem to solve, computing all the necessary intermediate steps
    my ($hits, $netBothWays, $nestingNet, $acc2mode2ds2t, $acc2mode2seqBol, $dt2E, $Emax) = @_;
    
    ### This block has a lot of pre-elimination processing, to structure problem in a way that makes solving the easiest
    # remember what overlaps are forbidden, and store the final scores to make a single hash call per pair.
    my $overlapNet = Hits::overlapNetBinarySearch($hits, $nestingNet);
    my $scoresNet = getScoresNet_nonOvNeg($hits, $netBothWays, $dt2E, $Emax, $overlapNet);
    
    # this means we have to run lp_solve, otherwise we'll just parse the solution that's already present
    # get the list of families/modes that are GA, since we know they will pass automatically (no need to check their sequence scores against the sequence threshold)
    # an assumption breaks down when using negative context, but I doubt that it will matter in practice, let's just trust GAs when it comes to the sequence threshold.
    my $hitsGa = Hits::filterHitsBooleanCol($hits, 'GA');
    my ($acc2pass, $acc2mode2pass) = getFamModeAutoPass($hitsGa);
    # also simplify thresholds structure for faster access across iterations
    my ($hit2td, $acc2mode2ts, $hit2pass, $hit2passFam) = getThresholdsPerHit($hits, $acc2mode2ds2t, $acc2pass, $acc2mode2pass, $acc2mode2seqBol);
    # this calculates the interpolated threshold for the current set of domains (in our lp_solve setup, we might have to iterate this calculation to make sure that we estimated the iteration correctly)
    my $acc2ts = getInterpolatedTs($hits, $acc2mode2ts, $hit2passFam);
    
    # calculate and write the lp_solve lines...
    my ($nodesEdgesDeclarations, $objectiveFunction, $edgeDefConstraints) = writeLpDeclarationsObjectiveEdges($hits, $hit2td, $scoresNet);
    my $overlapConstraints = writeLpOverlaps($hits, $overlapNet);
    my ($seqDeclarations, $threshDomConstraints, $threshSeqConstraints, $seqDefConstraints) = writeLpThresholdsFamDeclarations($hits, $scoresNet, $hit2td, $acc2ts, $hit2pass, $hit2passFam);
    # this next scalar contains the entire problem as text (all lines concatenated), which we will pipe to problem instead of printing to a file first
    my $lp = writeLpFile($nodesEdgesDeclarations, $objectiveFunction, $edgeDefConstraints, $overlapConstraints, $seqDeclarations, $threshDomConstraints, $threshSeqConstraints, $seqDefConstraints);
    
    # this is all we wanted, return text of lp_solve problem
    # actually, we reuse scoresNet later, return it too
    return ($lp, $scoresNet);
}

# useful for anything that uses lp_solve, but so far only used for nonOvNeg
sub readLpSol {
    # reads the output text rather than reading the contents of a file, since now data is all piped and not written to files
    # all we need to do is pass the hits with the indeces that passed according to the output
    # a previous version used to interpret failure and return a simplified solution, but this version simply returns empty if solution is empty, regardless of the reason
    # probably not the most efficient, but easy to understand code.  Files shouldn't be so big that this matters.
    my ($sol, $hits) = @_;
    # first split input text by lines
    my @lines = split /\n/, $sol;
    # the first thing we want to know is the outcome of the job, which is the third line (1-based).
    # the only outcome that requires no editing so far is Timeout
    my $outcome = $lines[2];
    $outcome = $mallocShortStr if $outcome =~ /^$mallocLongStr/o;
    $outcome = $suboptimalStr if $outcome =~ /^$suboptimalStr/o;
    $outcome = $optimalStr unless $outcome;
    # grep to get the solution domains
    # we only care about domains, not edges or families, and only if their value is 1
    @lines = grep {/^D\d+\s+1$/} @lines;
    # now actually get indeces and transfer hits
    my @hitsPass;
    foreach (@lines) {
	my ($i) = /^D(\d+)/;
	push @hitsPass, $hits->[$i];
    }
    # I'm not sure why, but they are not sorted by start position
    return (Hits::sortHitsByCol(\@hitsPass, 'start'), $outcome);
}

# all the subs called writeLp are used for nonOvNeg only, so far, but in theory they could be used for posElim too!!! (if we want to let lp_solve handle the positive elimination instead of us directly!)
sub writeLpOverlaps {
    # this writes the constraints for the disallowed overlaps (precalculated)
    my ($hits, $overlapNet) = @_;
    my $overlapConstraints = ''; # the lines we'll return
    my $numHits = scalar @$hits;
    for (my $i = 0; $i < $numHits; $i++) {
	my $hi = $hits->[$i];
	my $overlapNetHere = $overlapNet->{$hi};
	for (my $j = $i+1; $j < $numHits; $j++) {
	    my $hj = $hits->[$j];
	    $overlapConstraints .= 'D'.$i.' + D'.$j." <= 1;\n" if $overlapNetHere->{$hj};
	}
    }
    return $overlapConstraints;
}

sub writeLpDeclarationsObjectiveEdges {
    # we want to write the objective function, but we get the edge-defining constraints and declarations trivially, so might as well get them
    # the scores net already excludes scores between overlapping domains, so we won't waste text writing that, and we won't define those edges that cannot be used!
    my ($hits, $hit2td, $scoresNet) = @_;
    # these are the lines we want
    my ($nodesEdgesDeclarations, $objectiveFunction, $edgeDefConstraints) = ('bin ','','');
    my $numHits = scalar @$hits;
    for (my $i = 0; $i < $numHits; $i++) {
	my $hi = $hits->[$i];
	# declare this domain
	$nodesEdgesDeclarations .= 'D'.$i.', ';
	# the "normalized self score" of this domain (Hi-Ti)
	$objectiveFunction .= ($hi->{score} - $hit2td->{$hi}).' D'.$i.' ';
	# move into pairs
	my $scoresNetHere = $scoresNet->{$hi};
	for (my $j = $i+1; $j < $numHits; $j++) {
	    my $hj = $hits->[$j];
	    my $Cij = $scoresNetHere->{$hj};
	    if ($Cij) {
		# there's a score, so we need an edge variable
		# declare the edge
		$nodesEdgesDeclarations .= 'E'.$i.'_'.$j.', ';
		# add to objective function
		$objectiveFunction .= (2*$Cij).' E'.$i.'_'.$j.' ';
		# write edge constraint, same form for all edges
		$edgeDefConstraints .= '0 <= D'.$i.' + D'.$j.' - 2 E'.$i.'_'.$j." <= 1;\n";
	    }
	}
    }
    # finish objective function line with semicolon and endline (replacing last space, just more aesthetically pleasing)
    $objectiveFunction =~ s/ $/;\n/;
    # also need to close declarations (replace last comma with semicolon and endline)
    $nodesEdgesDeclarations =~ s/, $/;\n/;
    # done! return!
    return ($nodesEdgesDeclarations, $objectiveFunction, $edgeDefConstraints);
}

sub writeLpThresholdsFamDeclarations {
    my ($hits, $scoresNet, $hit2td, $acc2ts, $hit2pass, $hit2passFam) = @_;
    # these are the lines we want
    my ($seqDeclarations, $threshDomConstraints, $threshSeqConstraints, $seqDefConstraints) = ('','','','');
    # this is a temporary hash that stores the equations for the whole family (the sum of $threshDomConstraints from the same family) that get collapsed into a single $threshSeqConstraints entry.
    my %acc2threshSeqConstraints;
    # we also need for each equation a constant with all the negative factors it contains (the problem is that, when a sequence threshold is not required for one mode but not the other, we want this constraint to be satisfied despite the presence of domains of the mode that passes, but this sum of scores might be negative so passing is not trivial [although very unlikely since each has to pass the family thresholds and they are not usually negative]).  We only need this constant from the domains that pass automatically.
    my %acc2threshSeqNeg;
    # this hash maps for each sequence indicator (accession/family) the (text) list of domains that define it (via an OR switch).
    my %acc2domIndicators;
    my $numHits = scalar @$hits;
    for (my $i = 0; $i < $numHits; $i++) {
	my $hi = $hits->[$i];
	my $scoresNetHere = $scoresNet->{$hi}; # copy down net for this domain
	# there's a common case (for at least one unique, large protein, PF10_0374, anyway) in which a domain has no context at all (positive or negative), so passing really only depends on the self score.  These are either GA (pass without context) or eliminated in the first round of posElim, so let's assume they're GA and they pass automatically, so we won't bother writing the equation.
	if ($scoresNetHere) {
	    my ($acc, $score) = @{$hi}{qw(acc score)}; # a complicated hash slice
	    my $pass = $hit2pass->{$hi}; # remember if this hit needs to pass the family threshold
	    my $passFam = $hit2passFam->{$hi}; # remember if this hit needs to be considered in the family threshold (although its mode might allow it to pass automatically)
	    # the "normalized self score" of this domain (Hi-Ti), for the domain threshold
	    $threshDomConstraints .= ($score - $hit2td->{$hi}).' D'.$i.' ';
	    unless ($passFam) {
		# we need an unnormalized score for the sequence threshold
		$acc2threshSeqConstraints{$acc} .= $score.' D'.$i.' ';
		# we'll use this score in the negative constant if the domain passes regardless of the sequence threshold (and if the score is negative, of course)
		if ($pass) { $acc2threshSeqNeg{$acc} += $score if $score < 0; }
		# else the domain indicator is needed to define the sequence indicator
		else { $acc2domIndicators{$acc} .= '+ D'.$i.' '; }
	    }
	    # move into pairs
	    for (my $j = 0; $j < $numHits; $j++) { # unlike objective function, here we go through all pairs! (even the ones before $i, but we need to exclude $i itself)
		next if $j == $i; # here we exclude $i
		my $hj = $hits->[$j];
		my $Cij = $scoresNetHere->{$hj};
		if ($Cij) {
		    # there's a score to add to the thresholds!
		    # first sort so i comes before j (otherwise edge indicator will have the wrong name!)
		    my $ij = ($i < $j) ? $i.'_'.$j : $j.'_'.$i;
		    # this is the term that we use in both our equations (precalculate stringification?)
		    my $term = $Cij.' E'.$ij.' ';
		    # add to domain threshold
		    $threshDomConstraints .= $term;
		    unless ($passFam) {
			# also add to sequence threshold if necessary
			$acc2threshSeqConstraints{$acc} .= $term;
			# we'll use this score in the negative constant if the domain passes regardless of the sequence threshold (and if the score is negative, of course)
			if ($pass && $Cij < 0) {
			    $acc2threshSeqNeg{$acc} += $Cij;
			}
		    }
		}
	    }
	    # we're done with this hit, close domain threshold
	    $threshDomConstraints .= ">= 0;\n";
	}
    }
    
    # lastly, navigate the sequence indicators (accessions/families) that need a sequence threshold and finish the sequence indicator definitions and the threshold itself
    while (my ($acc, $threshSeqConstraint) = each %acc2threshSeqConstraints) {
	# part 0: declare sequence indicator
	$seqDeclarations .= $acc.', ';
	
	# part I: sequence threshold constraint
	# get the pre-calculated interpolated sequence threshold
	my $ts = $acc2ts->{$acc};
	# get the negative constant, if present, for automatic passing of one mode if one needs the threshold but not the other mode
	my $threshSeqNeg = $acc2threshSeqNeg{$acc};
	# finish the threshold constraint using this threshold, the sequence indicator, and the negative constant if present
	if ($threshSeqNeg) { $threshSeqConstraint .= ($threshSeqNeg - $ts).' '.$acc.' '.(-$threshSeqNeg); }
	# this is the easy case in which there is no constant to worry about
	else { $threshSeqConstraint .= (-$ts).' '.$acc; }
	# in both cases, close the constraint and copy it to the main list
	$threshSeqConstraints .= $threshSeqConstraint." >= 0;\n";
	
	# part II: sequence indicator definition constraint
	# first get list of domains that define this sequence indicator
	my $domIndicators = $acc2domIndicators{$acc};
	# then get the number of domains in this list (via a perlish hack, count the number of D's)
	my $n = $domIndicators =~ tr/D/D/;
	# now we write the two equations that define the sequence indicator, and copy them to the main list
	$seqDefConstraints .= $domIndicators.' -'.$n.' '.$acc." <= 0;\n"; # so sequence indicator is 1 if at least one of the domain indicators is 1
	$seqDefConstraints .= $domIndicators.' -'.$acc." >= 0;\n"; # so sequence indicator isn't 1 if none of the domain indicators are 1
    }
    # if there are declarations, we need to close them (replace last comma with semicolon and endline, start line with "bin")
    if ($seqDeclarations) {
	$seqDeclarations =~ s/, $/;\n/;
	$seqDeclarations = 'bin '.$seqDeclarations;
    }
    
    # done! return!
    return ($seqDeclarations, $threshDomConstraints, $threshSeqConstraints, $seqDefConstraints);
}

sub writeLpFile {
    # this version just returns the string with the problem we want to run, doesn't write to any files
    my ($nodesEdgesDeclarations, $objectiveFunction, $edgeDefConstraints, $overlapConstraints, $seqDeclarations, $threshDomConstraints, $threshSeqConstraints, $seqDefConstraints) = @_;
    
    # print to file
    return join '',
    "// Objective function (sum of scores)\n",
    $objectiveFunction,
    "// Edge-defining constraints (AND of nodes Di and Dj, wish this was more trivial)\n",
    $edgeDefConstraints,
    "// Sequence indicator-defining constraints (OR of list of nodes Di in family of sequence, wish this was more trivial)\n",
    $seqDefConstraints,
    "// Enforce Pfam domain thresholds constraints\n",
    $threshDomConstraints,
    "// Enforce Pfam sequence thresholds constraints (only when they're necessary, rarely the case)\n",
    $threshSeqConstraints,
    "// Constraints to force domains to not overlap\n",
    $overlapConstraints,
    "// Declarations (we need to say all nodes, edges, and family indicators are binary)\n",
    $nodesEdgesDeclarations,
    $seqDeclarations,
    ;
}

sub posElim {
    # this sub does everything that must be done for posElim, while receiving as inputs all that can be factored out, and the problem itself ($hits)
    my ($hits, $netBothWays, $nestingNet, $acc2mode2ds2t, $acc2mode2seqBol, $Ecut) = @_;
    
    # start basic analysis
    # total number of hits
    my $numHits = scalar @$hits;
    
    # separate the domains that already passed thresholds (incorrectly called GA), and the ones that don't split into domains that have context partners and things that just don't pass
    # we may also place an E-value cutoff on the @$hitsPass domains, to ensure quality of prospectives (but no such threshold applies to @$hitsGa!)
    my ($hitsGa, $hitsPass) = separateHitsByType($hits, $netBothWays, $Ecut);
    # count components
    my $numGa = scalar @$hitsGa;
    my $numPass = scalar @$hitsPass;
    my $numNonCc = $numHits - $numGa - $numPass;
    
    ### This block has a lot of pre-elimination processing, to structure problem in a way that makes solving the easiest
    # now that we eliminated things that just don't have context, we can remember what overlaps are forbidden (instead of recomputing them at every iteration below), and storing the final scores to make a single hash call per pair.
    my $hitsAll = [@$hitsGa, @$hitsPass]; # the total set of domains, to get scores and thresholds that count everything
    # list of indeces, to keep track of eliminated domains without altering most of the structures, and to avoid hash mappings (do array mappings instead).
    my $hitisGa = [0..$numGa-1];
    my $hitisPass = [$numGa..$numGa+$numPass-1];
    # an expensive part of getScoresNet_posElim is mapping hits to their accessions, so we can get their scores
    # let's do that mapping in O(n) instead of in the O(n^2) loop
    my @hitisAll2accInt;
    for (my $i = 0; $i < $numGa+$numPass; $i++) {
	my $acc = $hitsAll->[$i]{acc}; # this does the simple mapping
	$acc =~ s/^PF//; # for Pfam, removes the PF label, so all that's left is a number
	$hitisAll2accInt[$i] = int $acc; # store in final structure, hopefully as a number internally
    }
    # now actually get scores
    getScoresNet_posElim(\@hitisAll2accInt);
    # get the list of families/modes that are GA, since we know they will pass automatically (no need to check their sequence scores against the sequence threshold)
    my ($acc2pass, $acc2mode2pass) = getFamModeAutoPass($hitsGa);
    # also simplify thresholds structure for faster access across iterations
    # NOTE: could do this before "resolveLsFs", but that might generate a hash with too many entries that are never used and which complicate searching.  I think it's faster like this (but I haven't tested it)
    my ($hit2td, $acc2mode2ts, $hit2pass, $hit2passFam) = getThresholdsPerHit($hitsAll, $acc2mode2ds2t, $acc2pass, $acc2mode2pass, $acc2mode2seqBol);
    
    # do massive eliminations by setting cutoffs at the domain and family level
    my $numNoPass = 0;
    my $iter = 1;
    $hitisPass = elimMaxScore($hitsAll, $hitisGa, $hitisPass, $hit2td, $acc2mode2ts, $hit2pass, $hit2passFam);
    my $numPassNow = scalar @$hitisPass;
    while ($numPass != $numPassNow) { # this condition means we haven't converged
	# update counts
	$numNoPass += $numPass-$numPassNow;
	$numPass = $numPassNow;
	# run again
	$iter++;
	$hitisPass = elimMaxScore($hitsAll, $hitisGa, $hitisPass, $hit2td, $acc2mode2ts, $hit2pass, $hit2passFam);
	$numPassNow = scalar @$hitisPass;
    }
    
    # now that total hits are largely reduced, include overlap information into eliminations (too expensive to do at first)
    my @hitisAllLeft = (@$hitisGa, @$hitisPass); # this list gets fed into the scoresNet updater
    my $hitsAllLeft = [@{$hitsAll}[@hitisAllLeft]]; # this hist list includes only GA and the hits that have passed.  This reduced list requires overlap checking.
    my $overlapNet = Hits::overlapNetBinarySearch($hitsAllLeft, $nestingNet); # get which hits actually overlap and aren't allowed to, mapping from reference
    zeroScoresOverlappingPairs($hitsAll, \@hitisAllLeft, $overlapNet); # lastly, update $scoresNet half matrix by zeroing the scores that aren't allowed due to overlaps
    
    # do massive eliminations by setting cutoffs at the domain and family level
    $iter++;
    $hitisPass = elimMaxScore($hitsAll, $hitisGa, $hitisPass, $hit2td, $acc2mode2ts, $hit2pass, $hit2passFam);
    $numPassNow = scalar @$hitisPass;
    while ($numPass != $numPassNow) { # this condition means we haven't converged
	# update counts
	$numNoPass += $numPass-$numPassNow;
	$numPass = $numPassNow;
	# run again
	$iter++;
	$hitisPass = elimMaxScore($hitsAll, $hitisGa, $hitisPass, $hit2td, $acc2mode2ts, $hit2pass, $hit2passFam);
	$numPassNow = scalar @$hitisPass;
    }
    
    # done, turn passed indeces into a finalized domain list, to return
    $hitsAll = [@{$hitsAll}[@$hitisGa, @$hitisPass]]; # this updates hitsAll to include only GA and the hits that have passed.  It's a very complicated mess of dereferences and re-references
    
    # done, we got the posElim solution to the input set of hits!  Also return some stats
    return ($hitsAll, $iter, $numHits, $numGa, $numPass, $numNoPass, $numNonCc);
}

# actually this had to be split due to $scoresNet differences (posElim uses double array, nonOvNeg uses double hash)
# since, wrapUp_posElim has been deprecated, so this is actually the only version that exists, but we have to note that the inputs are nonOvNeg specific!
sub wrapUp_nonOvNeg {
    my ($hits, $scoresNet) = @_;
    # calculate all context scores
    my $hit2contextScore = getSumScoreEdgesFromNode_nonOvNeg($hits, $scoresNet);
    # now get total domain and family scores, as well as HMM fam scores (cause we don't trust the ones in the structure, might not hold after eliminations)
    my %hit2scoreDom;
    my %acc2scoreSeq;
    my %acc2scoreHmmSeq;
    foreach my $h (@$hits) {
	my $contextScore = $hit2contextScore->{$h}; # may be undefined if sum was zero
	my ($scoreDom, $acc) = @{$h}{qw(score acc)};
	$acc2scoreHmmSeq{$acc} += $scoreDom;
	$scoreDom += 0.5*$contextScore if $contextScore; # for this case, the edge scores have to be halved!
	$hit2scoreDom{$h} = $scoreDom;
	$acc2scoreSeq{$acc} += $scoreDom;
    }
    # now make new hits with the modified scores (original hit scores will not be touched!)
    my @hitsNew;
    foreach my $h (@$hits) {
	my $acc = $h->{acc};
	# get total score
	my $scoreTotalDom = sprintf '%0.1f', $hit2scoreDom{$h};
	my $scoreTotalSeq = sprintf '%0.1f', $acc2scoreSeq{$acc};
	# get HMM score only (old score)
	my $scoreHmmDom = $h->{score};
	my $scoreHmmSeq = sprintf '%0.1f', $acc2scoreHmmSeq{$acc};
	# infer the context score
	my $scoreContextDom = sprintf '%0.1f', $scoreTotalDom-$scoreHmmDom;
	my $scoreContextSeq = sprintf '%0.1f', $scoreTotalSeq-$scoreHmmSeq;
	# now store these numbers again.  Note score and scoreSeq will now refer to the total scores
	my %hNew = %$h; # this should copy the hits, so we can edit them without changing the originals!
	$hNew{score} = $scoreTotalDom;
	$hNew{scoreSeq} = $scoreTotalSeq;
	$hNew{scoreHmm} = $scoreHmmDom;
	$hNew{scoreHmmSeq} = $scoreHmmSeq;
	$hNew{scoreContext} = $scoreContextDom;
	$hNew{scoreContextSeq} = $scoreContextSeq;
	push @hitsNew, \%hNew;
    }
    # return new hits with modified scores
    return \@hitsNew;
}

# this is for posElim only (eliminations works with the positive scores assumption, and besides nonOvNeg only eliminates through lp_solve, outside of Perl)
sub elimMaxScore {
    my ($hitsAll, $hitisGa, $hitisPassPrevious, $hit2td, $acc2mode2ts, $hit2pass, $hit2passFam) = @_;
    # go through hitsPassPrevious and separate those that have a partner from those that don't, and those whose partners may raise their scores enough to pass the threshold
    my @hitisAll = (@$hitisGa, @$hitisPassPrevious); # the total set of domains, to get scores that count everything
    my $numHitisGa = scalar @$hitisGa;
    my $numHitisAll = scalar @hitisAll;
    # calculate all context scores
    my $hitk2contextScore = getSumScoreEdgesFromNode_posElim(\@hitisAll);
    # now get sequence scores, and count the ls and fs domains to get ratio of ls to fs members of the family
    my %acc2scoreSeq;
    # calculate the sum of scores for each family that needs the interpolated thresholds
    for (my $k = 0; $k < $numHitisAll; $k++) {
	# for now we need to get the original hit structure, to use the hit2blah mappings, but we should consider permanently shifting into indeces for all code!
	my $h = $hitsAll->[$hitisAll[$k]];
	# decide whether a family needs a sequence score by looking at the passFam boolean (family may pass automatically, so why bother with the sequence score)
	next if $hit2passFam->{$h};
	# get all these quantities of the hit at the same time (is it faster?)
	my ($score, $acc) = @{$h}{qw(score acc)}; # weird hash slice
	$acc2scoreSeq{$acc} += $h->{score} + $hitk2contextScore->[$k]/$scoreScale; # Need to reverse context score int scaling!
    }
    # now for each family that didn't pass automatically, remember if the sequence score passed this threshold or not.  (Remember that some modes pass automatically, even if this interpolated threshold doesn't pass, so we need to watch out!)
    # this passing is true for all members of the family, up to mode, so it makes sense to calculate it separately from the individual domain thresholds (this way we're factoring that calculation out!)
    # calculate the interpolated sequence thresholds we need
    my $acc2ts = getInterpolatedTs($hitsAll, $acc2mode2ts, $hit2passFam);
    my %acc2passFam; # remember if the sequence score passed the interpolated sequence threshold
    while (my ($acc, $scoreSeq) = each %acc2scoreSeq) {
	# get precalculated interpolated sequence threshold
	my $ts = $acc2ts->{$acc};
	# now we check if the threshold is passed by the sequence score
	$acc2passFam{$acc} = 1 if $scoreSeq >= $ts;
    }
    # now browse the domains (that are not GA), decide whether they still pass or not.  In terms of $k (the index of $i), we start after the last GA domain
    my @hitisPass; # has partners and max score allows domain to pass threshold
    for (my $k = $numHitisGa; $k < $numHitisAll; $k++) {
	my $contextScore = $hitk2contextScore->[$k];
	# if this had no partners, no point in looking at thresholds (since this isn't already GA, then it automatically doesn't pass)
	if ($contextScore) {
	    # now we're actually interested in what $i is...
	    my $i = $hitisAll[$k];
	    # for now we need to get the original hit structure, to use the hit2blah mappings, but we should consider permanently shifting into indeces for all code!
	    my $h = $hitsAll->[$i];
	    # test if it passed the threshold here or not
	    my $scoreDom = $h->{score} + $contextScore/$scoreScale; # add self HMM score and max context score, reverse context score int scaling
	    if ($scoreDom >= $hit2td->{$h}) { # compare to domain threshold
		# domain threshold was passed, now check sequence threshold
		# explaining the two following parts:
		# the first is true if the domain automatically passes the sequence threshold (due to many reasons).  This includes domains that pass the threshold for their particular mode, so it might be true even if the next isn't (they are not required to pass the interpolated sequence threshold).
		# the second is true if the family of this domain passed the interpolated threshold (but this was only evaluated for domains that didn't pass automatically, so we need to put it second)
		if ($hit2pass->{$h} || $acc2passFam{$h->{acc}}) {
		    push @hitisPass, $i; # this passed
		}
	    }
	    # else max score wasn't good enough
	}
    }
    return \@hitisPass;
}

# for both posElim and nonOvNeg
sub getInterpolatedTs {
    # $hits is hitsAll!
    my ($hits, $acc2mode2ts, $hit2passFam) = @_;
    # count the ls and fs domains to get ratio of ls to fs members of the family
    my %acc2mode2num;
    foreach my $h (@$hits) {
	# decide whether a family needs a sequence threshold by looking at the passFam boolean (family may pass automatically, so why bother with the sequence threshold)
	next if $hit2passFam->{$h};
	# get all these quantities of the hit at the same time (is it faster?)
	my ($acc, $mode, $ga) = @{$h}{qw(acc mode GA)}; # weird hash slice
	$acc2mode2num{$acc}{$mode}++; # tally how many members of each mode we have so we can interpolate the threshold accordingly
    }
    # now for each family that didn't pass automatically, calculate the ratio of ls to fs and compute the interpolated family threshold, then remember if the sequence score passed this threshold or not.  (Remember that some modes pass automatically, even if this interpolated threshold doesn't pass, so we need to watch out!)
    # this passing is true for all members of the family, up to mode, so it makes sense to calculate it separately from the individual domain thresholds (this way we're factoring that calculation out!)
    my %acc2ts; # the interpolated sequence thresholds
    # always start nagivating the full list.  In the case of posElim, most domains will only have the thresholds in that list anyway
    while (my ($acc, $mode2num) = each %acc2mode2num) {
	# get the sequence thresholds for each mode (a weird hash slice).
	my ($tsLs, $tsFs) = @{$acc2mode2ts->{$acc}}{qw(ls fs)};
	# a weird hash slice that gives us the number of ls and fs members for this accession only
	my ($numLs, $numFs) = @{$mode2num}{qw(ls fs)};
	# compute interpolated sequence threshold
	my $ts;
	# special cases first (presumably the most common ones)
	# these are the same no matter what kind of mean (arithmetic, geometric, harmonic) we use
	# since we have a member of the family, one being zero implies that the other is non-zero
	# in this case there are no ls, so the threshold should be from fs
	if (!$numLs) { $ts = $tsFs; }
	# opposite case (no fs, so threshold is from ls)
	elsif (!$numFs) { $ts = $tsLs; }
	# interpolation case (members from both families)
	# the previous checking guarantees that we're not using undefined values (perl would complain in the sums and multiplications)
	# note that we're not using the normal means on the numbers because the generalized means only work on positive numbers, while our thresholds may be zero or negative.  We've devised a way to apply the generalized means to a normalized form of the problem (essentially set the min to 1, and the max to 2), then unnormalize to get the actual values we want.  The equations behave well and hopefully they solve our problems...
	else {
	    # we need to sort the thresholds, to know which one is the min (n) and which one is the max (m)
	    my ($Tm, $Tn, $m, $n);
	    if ($tsLs > $tsFs) { ($Tm, $Tn, $m, $n) = ($tsLs, $tsFs, $numLs, $numFs); }
	    else { ($Tm, $Tn, $m, $n) = ($tsFs, $tsLs, $numFs, $numLs); }
	    # compute the ratio of max to total counts, this is the "natural" weight for the max (and 1-$lambda for the min), or the one that we would use for the arithmetic mean.
	    my $lambda = $m/($m + $n);
	    # now perform the interpolation with the final lambda
	    $ts = $Tm*$lambda + $Tn*(1-$lambda);
	}
	# store interpolated sequence threshold
	$acc2ts{$acc} = $ts;
    }
    return \%acc2ts;
}

# this is shared by both posElim and nonOvNeg
sub getThresholdsPerHit {
    # the idea is to get thresholds once from the large Pfam triple hash (each hit only has one Ti and one Tfi), and to put them in a hash that maps the hit (not the accession or anything hard to get) to the thresholds
    # it also makes a shorter version of the input hash, that only contains the accessions in use in the current protein, as opposed to the entire Pfam database, and only for the families that aren't GA (in both modes) and which don't have a necessary sequence threshold (in both modes too).
    my ($hits, $acc2mode2ds2t, $acc2pass, $acc2mode2pass, $acc2mode2seqBol) = @_;
    
    # first we do a survey of what modes and accessions we have, to be able to assess better what passes automatically and what doesn't
    # we also store their thresholds
    my %hit2td; # maps to the domain threshold
    my %acc2modes;
    my %acc2c; # counts members per family, we really only want to know which families have a single member so we only apply one of the thresholds (max td,ts) and set the passFam as 1 later on
    foreach my $h (@$hits) {
	my ($acc, $mode) = @{$h}{qw(acc mode)}; # weird hash slice
	# this is annoying, but I have to make sure that the reference exists before sending it, otherwise I'm just sending a scalar and the changes aren't made
	General::add(\@{$acc2modes{$acc}}, [$mode]);
	# get domain threshold, store in new hash
	$hit2td{$h} = $acc2mode2ds2t->{$acc}{$mode}{d};
	# count for families
	$acc2c{$acc}++;
    }
    # now let's decide which families automatically pass the sequence thresholds, based on whether the thresholds imply it ($acc2mode2seqBol), or the GA data force it ($acc2pass, $acc2mode2pass), and lastly we can decide automatic passing in the subcase in which only one of the two modes are present (%acc2modes).  This subcase is based on the starting data, so if we eliminate members of a family such that only mode mode is left later on, we will not be able to use that extra data (since recomputing this might be too expensive to make it worth it).  Also, individual domains may pass automatically from a single mode, in the case that there are two modes present and one doesn't pass automatically, so this data structure will allow us to decide whether a sequence score needs to be computed using all the members of the family, for the sake of the domains that don't pass automatically.
    my %acc2passFam; # this enhances the input $acc2pass since it uses more info than just GA alone, but is only used internally
    while (my ($acc, $modes) = each %acc2modes) {
	my $familyPasses = 0;
	# this tells us whether the accession needs sequence thresholds since they're different from domain thresholds for both modes
	my $mode2seqBol = $acc2mode2seqBol->{$acc}; # this may be undefined, we also use it a few times so it makes sense to save it
	# this is the easy case, in which passing is for both modes
	if ($acc2pass->{$acc} || !$mode2seqBol) { $familyPasses = 1; }
	# the hardest case is when there are two modes and they each pass for different reasons (i.e. one is GA, the other doesn't have a different Ts).  This happens with some frequency
	# this generalizes the slightly less hard case, in which the presence of a single mode settles passing for the whole family or not
	else {
	    my $numModes = scalar @$modes;
	    my $numModesPass = 0;
	    foreach my $mode (@$modes) {
		# note the first IF forces $mode2seqBol to be true here (so it's a reference)
		# either it's GA or threshold is not required!
		if ($acc2mode2pass->{$acc}{$mode} || !$mode2seqBol->{$mode}) {
		    $numModesPass++;
		}		
	    }
	    $familyPasses = 1 if $numModesPass == $numModes;
	}
	# now store in a hash if the family automatically passes
	$acc2passFam{$acc} = $familyPasses if $familyPasses;
    }
    # let's clean the %acc2c data to include only the accs that occur exactly once (should be faster than doing the evaluation for each hit if there are large families)
    # after this, hash will return true only if accession occurred once
    while (my ($acc, $c) = each %acc2c) {
	delete $acc2c{$acc} unless $c == 1;
    }
    # lastly, let's permanently map this info to the hits themselves (whether they pass automatically as a family, or whether they pass individually, which implies that they are needed to compute the sequence thresholds anyway)
    my %hit2pass; # they pass sequence threshold individually (mode distinction)
    my %hit2passFam; # they pass sequence threshold as a family (no mode distinction, no need to calculate sequence score)
    foreach my $h (@$hits) {
	my ($acc, $mode) = @{$h}{qw(acc mode)}; # weird hash slice
	# easy case, whole family passes
	if ($acc2passFam{$acc}) {
	    $hit2pass{$h} = 1;
	    $hit2passFam{$h} = 1;
	}
	# in this case there's a single member, so "whole" family pases too but we need to transfer the max threshold (of ts,td) to td.
	elsif ($acc2c{$acc}) { # this implies $c == 1, look back a few lines to see that other entries were deleted
	    $hit2pass{$h} = 1;
	    $hit2passFam{$h} = 1;
	    # get alternative threshold (ts, sequence threshold, in a way td is the default so we only need to retrieve the ts)
	    my $ts = $acc2mode2ds2t->{$acc}{$mode}{s};
	    # change if the max is different
	    $hit2td{$h} = $ts if $ts > $hit2td{$h};
	}
	# hard case, only a single mode may pass automatically (or maybe neither)
	# evaluation mimics the single-mode hard case in the loop before, but here it's done on many more cases (when both modes are present)
	else {
	    my $mode2seqBol = $acc2mode2seqBol->{$acc}; # this may be undefined, we also use it a few times so it makes sense to save it
	    # the hash ref call is guaranteed to work since $acc2pass{$acc} would have returned true if $mode2seqBol was undefined
	    if ($acc2mode2pass->{$acc}{$mode} || !$mode2seqBol->{$mode}) {
		$hit2pass{$h} = 1;
	    }
	}
    }
    # this makes the short hash we wanted
    my %acc2mode2ts; # maps the family to the two family thresholds, for each mode (for interpolation), only bother for families that don't automatically pass
    foreach my $acc (keys %acc2modes) {
	unless ($acc2passFam{$acc}) { # this is for families that don't automatically pass
	    my %mode2ts; # the hash we want, we need to take out the domain threshold from the big structure
	    # navigate the big hash to get what we want
	    while (my ($mode, $ds2t) = each %{$acc2mode2ds2t->{$acc}}) {
		$mode2ts{$mode} = $ds2t->{s};
	    }
	    # store in our main structure
	    $acc2mode2ts{$acc} = \%mode2ts;
	}
    }
    return (\%hit2td, \%acc2mode2ts, \%hit2pass, \%hit2passFam);
}

# this is shared by both posElim and nonOvNeg
sub getFamModeAutoPass {
    # the families that are GA for mode m automatically pass the sequence threshold for mode m, since they can only get larger with more hits (context is only positive, and hits are only positive after the first one [this is actually a weird HMMer2 side effect, will not be the case with HMMer 3]!)
    # an assumption breaks down when using negative context, but I doubt that it will matter in practice, let's just trust GAs when it comes to the sequence threshold.
    # we also want to make a shorter list of families that pass for both modes, this is likely to be the case for some families
    my ($hitsGa) = @_;
    my (%acc2pass, %acc2mode2pass);
    # first pass stores the passing per acc/mode
    foreach my $h (@$hitsGa) {
	my ($acc, $mode) = @{$h}{qw(acc mode)}; # weird hash slice
	$acc2mode2pass{$acc}{$mode} = 1;
    }
    # second pass stores the accessions that passed for both modes
    while (my ($acc, $mode2pass) = each %acc2mode2pass) {
	my $passSum = 0;
	foreach my $pass (values %$mode2pass) {
	    $passSum += $pass;
	}
	$acc2pass{$acc} = 1 if $passSum == 2; # accession passed both modes if the sum is 2
    }
    return (\%acc2pass, \%acc2mode2pass);
}

# the only reason we have a nonOvNeg version is that this takes the scoresNet that is a double hash, too painful to rewrite the rest of nonOvNeg to use the double array system that posElim uses.  This is used specifically for the wrapUp_nonOvNeg, and nothing else.
sub getSumScoreEdgesFromNode_nonOvNeg {
    # returns a hash that maps each hit to its total context score (should be halved in some contexts since edges are shared)
    my ($hits, $scoresNet) = @_;
    my %hit2score; # this is the hash we want
    my ($h, $h2, $scoresNetThisHit, $sumContextScore, $edgeScore);
    foreach $h (@$hits) {
	$scoresNetThisHit = $scoresNet->{$h}; # copy down structure, should reduce the number of hash calls
	$sumContextScore = 0; # this is the context score it gets from all possible interactions
	foreach $h2 (@$hits) {
	    $edgeScore = $scoresNetThisHit->{$h2}; # score of this pair
	    $sumContextScore += $edgeScore if defined $edgeScore; # will be undefined if this wasn't a context pair
	}
	$hit2score{$h} = $sumContextScore if $sumContextScore; # store the total score if non-zero (saves a bit of memory and hash lookup time?)
    }
    return \%hit2score;
}


# this is the nonOvNeg version (uses negative scores, special rules apply to them!).
# A lot of the code is shared with *_posElim, but the tighter integration of each for their particular needs is worth for the speed differences I anticipate.
sub getScoresNet_nonOvNeg {
    # this sub combines the overlap disallowed rule, with the context net that uses accessions, to make a net that maps each hit pair (as refs, not Pfam accessions) to the scores they should map to.  This should significantly speed up iterations (without affecting single-iteration runs a whole lot)
    # providing an empty overlap hash should have no effect in processing, I only hope perl doesn't complain
    my ($hits, $netBothWays, $dt2E, $Emax, $overlapNet) = @_;
    my %scoresNet; # the structure we want, both ways
    # we can halve the number of operations by not repeating pairs in loop!
    my $numHits = scalar @$hits;
    for (my $i = 0; $i < $numHits; $i++) {
	my $hi = $hits->[$i];
	my $acci = $hi->{acc}; ### in nonOvNeg this is reused
	my $netThisAcc = $netBothWays->{$acci}; # copy down structure, should reduce the number of hash calls
	for (my $j = $i+1; $j < $numHits; $j++) {
	    my $hj = $hits->[$j];
	    my $accj = $hj->{acc}; ### in nonOvNeg this is reused
	    my $edgeScore = $netThisAcc->{$accj}; # score of this pair
	    ### This is where the posElim-nonOvNeg difference lies
	    if (defined $edgeScore) {
		# this means we had non-zero counts.
		# However, we set to zero if we got a negative score.  This way we only have zero or positive scores when we have seen context.
		$edgeScore = 0 if $edgeScore < 0;
	    }
	    else {
		# this means we'll need to calculate the negative score
		# try to get energies, but not all domains have these
		my $Ei = $dt2E->{$acci};
		my $Ej = $dt2E->{$accj};
		# if we got both energies, continue with calculation
		if (defined $Ei && defined $Ej) {
		    $edgeScore = $Ei + $Ej - $Emax;
		    # now set to zero if we got a positive score.  This way we only have zero or negative scores when we haven't seen context.
		    $edgeScore = 0 if $edgeScore > 0;
		}
		# if one or both energies were undefined, we must be dealing with domains that were not observed in our filtered architectures (with more than one instance and more than one domain).  These domains should not have context with anything (they are esentially ignored context-wise, and should only be part of the problem in the first place if they are GA).
		else { $edgeScore = 0; }
	    }
	    ### end posElim-nonOvNeg difference
	    unless (!$edgeScore || $overlapNet->{$hi}{$hj}) {
		# score may be undefined (or zero), we check for that before bothering with storing it
		# then check for overlap and ask the nesting net if it should be allowed (precalculated)
		# if they pass, save both ways!
		$scoresNet{$hi}{$hj} = $edgeScore;
		$scoresNet{$hj}{$hi} = $edgeScore;
	    }
	}
    }
    return \%scoresNet;
}

# for posElim only, since it uses the half matrix $scoresNet (not the double hash net that nonOvNeg uses!)
sub zeroScoresOverlappingPairs {
    # this sub updates the $scoresNet half matrix by zeroing the scores that aren't allowed due to overlaps
    my ($hitsAll, $hitisAllLeft, $overlapNet) = @_;
    
    # C-like initialize all vars outside of potentially tight loop
    my ($i,$j,$k,$l,$ti) = (0,0,0,0,0);
    my ($hi, $hj) = ({},{});
    
    # browse the @$hitisAllLeft pairs with triangle indeces of indeces
    my $numHitisAllLeft = scalar @$hitisAllLeft;
    for ($k = 1; $k < $numHitisAllLeft; $k++) { # starts at k=1 cause l<k, so k=0 doesn't have an l!
	$i = $hitisAllLeft->[$k];
	$hi = $hitsAll->[$i];
	for ($l = 0; $l < $k; $l++) {
	    $j = $hitisAllLeft->[$l];
	    $hj = $hitsAll->[$j];
	    # zero score if we have a disallowed overlap
	    if ($overlapNet->{$hi}{$hj}) { # works cause net is both ways
		# zero score by calling the C function, which interacts with the C data directly
		zeroScore(General::t($i, $j));
	    }
	}
    }
    # done, nothing to return since $scoresNet was altered by reference
}

# this is only used in posElim (but might be used in nonOvNeg if we wanted to skip posElim altogether)
sub separateHitsByType {
    # separate the domains that already passed thresholds (called GA here), and the ones that don't split into domains that have context partners and things that just don't pass
    # we may also place an E-value cutoff on the @$hitsPass domains, to ensure quality of prospectives (but no such threshold applies to @$hitsGa!)
    my ($hits, $netBothWays, $Ecut) = @_;
    my $isEcut = defined $Ecut; # boolean that remembers if Ecut was defined or not
    my @hitsGa;
    my @hitsPass;
    my %accPass2c;
    foreach my $h (@$hits) {
	my $acc = $h->{acc};
	# GA pass automatically, have their own list
	if ($h->{GA}) {
	    push @hitsGa, $h;
	    $accPass2c{$acc}++;
	}
	# this makes sure we don't add accessions to our list, and they don't pass, if they don't satisfy the Ecut, if defined
	elsif (!$isEcut || $h->{E} <= $Ecut) {
	    # we query whether an accession might carry context by sending it to C, as an int
	    my $accInt = $acc;
	    $accInt =~ s/^PF//; # should leave the int behind only
	    if (accIntIsCc($accInt)) {
		push @hitsPass, $h;
		$accPass2c{$acc}++;
	    }
	}
	# else didn't pass
    }
    # time for second round.  We know what accessions are present, now we'll only let pass domains that actually have partners present!
    # very fast elimination that is O(n), so we don't build that huge scores matrix and eliminate then, which is O(n^2)
    # as implemented, this approach works for pairs of different accessions, but will let things pass if they have no partners but have self context
    # (we would have to keep track of counts to realize that accs with self context need more than two counts of that acc to avoid being eliminated, but I won't worry about that for now)
    # for this to work, @accsPass needs to be sorted
    my @accsPass = sort keys %accPass2c;
    my @hitsPass2; # things that pass round 2
    foreach my $h (@hitsPass) {
	# get possible context pairs
	my @pairs = keys %{$netBothWays->{$h->{acc}}};
	# pass if any of the possible pairs is present in the pass list from round 1
	if (General::isIntersectionBinarySearch(\@pairs, \@accsPass)) {
	    push @hitsPass2, $h;
	}
    }
    return (\@hitsGa, \@hitsPass2);
}


1;
