# Copyright (C) 2010 Alejandro Ochoa, Singh Research Group, Princeton University
# This file is part of dPUC.
# dPUC is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# dPUC is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with dPUC.  If not, see <http://www.gnu.org/licenses/>.

package ContextDaemon;
use Time::HiRes ();
use lib '.';
use General;
use FilePaths;
use Hits;
use Context;
use strict;

# the purpose of this package is to load the fixed data at startup, and keep it in memory.
# in a single Perl command-line run, this is equivalent to loading the data to memory for a single session.
# in a mod_perl (Apache) run, this is equivalent to loading the data for all sessions once, factoring out that step.  Data will only be reloaded upon Apache restart (service httpd restart)

# this is the only package that uses FilePaths, but some of the other scripts use it too

# constants
my $isPosElim = 1; # a named constant
my $format = 'PFAM'; # defines the header of the output files

# get header for output format
our $header = $Hits::headers{$format};

# hashes to map accessions to simple properties (name, ls/fs ordering rule, clan membership)
my $acc2name = General::readHashFile($FilePaths::pfamName);
my $acc2order = General::readHashFile($FilePaths::pfamOrder); # Order of scoring (globalfirst, localfirst, byscore)
my $acc2clan = General::readHashFile($FilePaths::pfamClans);

# get Pfam thresholds, process them into two handy structures
my ($mode2acc2ds2ga, $acc2mode2ds2t) = Hits::getPfamThresholds($FilePaths::pfamGa);
# for posElim, get this boolean structure that tells us when sequence thresholds are necessary
my $acc2mode2seqBol = Hits::getPfamThresholdsSeqBol($acc2mode2ds2t);

# get pairs of accessions that are allowed to overlap (because they may be nested)
my $nestingNet = Hits::getNestingNet($FilePaths::pfamNesting);

# this wrapper choses the right file, given the parameters, and does all necessary processing
my $netBothWaysPosElim = Context::loadContextNet($isPosElim, $FilePaths::dt2dt2S);
my $netBothWaysNonOvNeg = Context::loadContextNet(undef, $FilePaths::dt2dt2S);

# for nonOvNeg, load accession "Energies" for calculating negative context scores.
my $dt2E = General::readHashFile($FilePaths::dt2E);
my $Emax = $dt2E->{max}; # a constant we use repeatedly

# subs

# this is here, in the Daemon, because it calls the above data structures as global variables.  It could be written so things are passed or called from package, but meh.
sub doSeqAll {
    # does all processing for a single sequence
    # also measures running times of the two main components
    # if provided with a filehandle (for log), prints progress of run
    
    # input parameters
    my ($id, $seq, $Ecut, $fhoLog) = @_;
    # note most parameters are global, no need to pass them every time
    
    # run HMMER, parse output into hits structures
    # $idLs and $idFs (the parsed protein names from HMMER) are ignored
    my $timeStart = Time::HiRes::time();
    print $fhoLog "Started Pfam LS run\n" if defined $fhoLog;
    my $outLs = Context::runHmmerPipe($id, $seq, $Ecut, $FilePaths::pfamLs, $FilePaths::hmmpfam);
    print $fhoLog "HMMER2 output was\n$outLs\n" if defined $fhoLog;
    my ($idLs, $hitsLs) = Hits::parseHmmerPfam($outLs, $acc2name, $mode2acc2ds2ga->{ls}, 'ls');
    print $fhoLog "Started Pfam FS run\n" if defined $fhoLog;
    my $outFs = Context::runHmmerPipe($id, $seq, $Ecut, $FilePaths::pfamFs, $FilePaths::hmmpfam);
    print $fhoLog "HMMER2 output was\n$outFs\n" if defined $fhoLog;
    my ($idFs, $hitsFs) = Hits::parseHmmerPfam($outFs, $acc2name, $mode2acc2ds2ga->{fs}, 'fs');
    # get Standard Pfam (GA = gathering thresholds)
    my $hitsGa = Hits::processHitsAsPfam($hitsLs, $hitsFs, $acc2order, $acc2clan);
    my $timeEnd = Time::HiRes::time();
    my $timeGa = $timeEnd - $timeStart; # count runtime
    print $fhoLog "Finished Standard Pfam processing, $timeGa\n" if defined $fhoLog;
    $hitsGa = Hits::deepCopyHits($hitsGa); # separate GA hits from dPUC hits, now we can modify hits without affecting Standard Pfam
    
    # now do all dPUC-specific processing
    $timeStart = Time::HiRes::time();
    # get consolidated ls-fs (resolveLsFs)
    my $hitsReLsFs = Hits::resolveLsFs($hitsLs, $hitsFs, $acc2mode2ds2t);
    print $fhoLog "Finished resolveLsFs\n" if defined $fhoLog;
    # run the dPUC positive elimination
    my ($hitsPosElim, $iter, $numHits, $numGa, $numPass, $numNoPass, $numNonCc) = Context::posElim($hitsReLsFs, $netBothWaysPosElim, $nestingNet, $acc2mode2ds2t, $acc2mode2seqBol, $Ecut);
    print $fhoLog "Finished Context posElim\n" if defined $fhoLog;
    # run the dPUC lp_solve general optimization
    my ($hitsNonOvNeg, $outcome) = Context::lpSolveNonOvNeg($hitsPosElim, $netBothWaysNonOvNeg, $nestingNet, $acc2mode2ds2t, $acc2mode2seqBol, $dt2E, $Emax, $FilePaths::timeout);
    print $fhoLog "Finished Context lpSolveNonOvNeg\n" if defined $fhoLog;
    $timeEnd = Time::HiRes::time();
    my $timeDpuc = $timeEnd - $timeStart; # count runtime
    
    # done, return the data of interest
    # main thing are hits (ours and Standard Pfam's), but we also have a bunch of stats about our context run and HMMER
    return ($hitsNonOvNeg, $hitsGa, $timeDpuc, $timeGa, $outcome, $iter, $numHits, $numGa, $numPass, $numNoPass, $numNonCc);
}

sub printOutputText {
    # prints predictions from Standard Pfam and dPUC, also stats of run
    my ($fho, $id, $Ecut, $hitsDpuc, $hitsGa, $timeDpuc, $timeGa, $outcome, $iter, $numHits, $numGa, $numPass, $numNoPass, $numNonCc) = @_;
    
    # truncate time significant figures, too many hurt my eyes
    $timeGa = sprintf '%0.4f', $timeGa;
    $timeDpuc = sprintf '%0.4f', $timeDpuc;
    
    # raw output is 0-based, end-exclusive.  Need to turn into a normal format for people (1-based, end-inclusive)
    $hitsGa = Hits::oneBasedEndInc($hitsGa);
    $hitsDpuc = Hits::oneBasedEndInc($hitsDpuc);
    
    # make different printing protein IDs for each prediction kind
    my $idGa = $id.'|Standard Pfam';
    my $idDpuc = $id.'|dPUC Pfam';
    
    # start printing domain predictions
    print $fho "# ".join("\t", @$header)."\n\n"; # print commented header for every protein, plus a blank space
    Hits::printHitsManyProts_OnTheFly($fho, $header, {$idGa => $hitsGa}); # Standard Pfam predictions
    print $fho "\n"; # separate entries
    Hits::printHitsManyProts_OnTheFly($fho, $header, {$idDpuc => $hitsDpuc}); # dPUC predictions
    print $fho qq{
# Runtime
#     Standard Pfam (including HMMER): $timeGa s
#     dPUC (on top of Standard Pfam): $timeDpuc s
# Positive elimination stats:
#     $Ecut E-value cutoff
#     $iter iterations
#     $numHits initial domains
#     $numGa Standard Pfam domains
#     $numPass candidate domains
#     $numNoPass domains with context that didn't pass thresholds
#     $numNonCc domains with no context
# Integer linear programming (lp_solve) stats:
#     "$outcome" outcome
//

};
    
}

1;
