# Copyright (C) 2010 Alejandro Ochoa, Singh Research Group, Princeton University
# This file is part of dPUC.
# dPUC is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# dPUC is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with dPUC.  If not, see <http://www.gnu.org/licenses/>.

use lib '.';
use General;
use FilePaths;
use Nets;
use strict;

# parameters
my $al = 1; # $al is the regularization parameter, which prevents scores from being infinite if there are zero counts.

# now actually get counts (single and pair, ...)
my ($u, $n, $dt2c, $dt2dt2c) = processPfamArch($FilePaths::pfamArchesExact);
my $ua = $u + $al; # the universe + pseudocount that we reuse througout the code

# compute background scores (will be used to compute negative context scores on the fly in other scripts)
my ($dt2E) = getFirstOrderP($dt2c, $n); # actually compute these probabilities
# add "max" energy, the constant in the negative context score computation (equivalent to Eij(cij == 0))
# this formula holds even for the uniform background
$dt2E->{max} = General::log2($n*$n*($u/$al + 1));
# print energies to hash file
General::printHashFile($FilePaths::dt2E, $dt2E);
# compute final pair scores for observed pairs
my ($dt2dt2S) = getSecondOrderPS($dt2dt2c, $dt2E, $al, $ua, $n);
# lastly, print pair scores as a net, for easy loading in other code
Nets::printNet($FilePaths::dt2dt2S, $dt2dt2S);

# subs

sub processPfamArch {
    # in this sub I confuse proteins with architectures often... same thing
    my ($fi) = @_;
    my $u = 0; # "size of universe", the number of domains in our database (restricted to multidomain architectures)
    my %dt2c; # domain types to counts of how many proteins have them at least once, used for first-order probabilities.
    my %dt2dt2c; # domain type pairs to counts, for both pairs and repeats
    # load architectures hash file
    my $arch2c = General::readHashFile($fi);
    while (my ($arch, $numInstances) = each %$arch2c) {
	next unless $numInstances > 1; # don't use this architecture if it only ocurred once in the database, high likelyhood it's a false positive
	my @accs = split /-/, $arch; # split architecture by dashes (-).  This is the domain list, but it contains repetitions...  Also note that PF prefix is missing!
	next unless @accs > 1; # skip proteins that have a single domain, they don't have pairs to count
	@accs = map { 'PF'.$_ } @accs; # this adds all the missing PF prefixes, now things are cool
	my $cp; # number of domains in this protein
	my %acc2c;
	foreach my $acc (@accs) {
	    $cp++; # one more domain in this protein
	    $acc2c{$acc}++; # here we get a non-redundant list of domains and their counts for this architecture/protein
	}
	$u += $cp*$numInstances; # add domains in this protein to the count of the universe
	@accs = sort keys %acc2c; # sorted non-redundant list of domains (overwrite that old, redundant array)
	my $numAccs = scalar @accs; # to set loops
	for (my $i = 0; $i < $numAccs; $i++) { # navigate non-redundant list of domains, in the right order so the domain network is right
	    my $acci = $accs[$i]; # get accession
	    my $ci = $acc2c{$acci}; # get count in this protein
	    $dt2c{$acci} += $ci*$numInstances; # add count of this domain on this protein to the total count of this domain on all proteins.  However, if using mona's scoring scheme, this should instead be this domain ratio calculated per protein (already a probability rather than a count, per protein).
	    for (my $j = $i; $j < $numAccs; $j++) { # navigate second domain, including first domain again to get repeats
		my $accj = $accs[$j]; # get accession
		my $cj = $acc2c{$accj}; # get count in this protein
		# count, by this point we're assured that this protein is multidomain, so no problem dividing by $cp - 1
		my $cij = ($i == $j) ? $ci*($ci-1) : $ci*$cj; # first count the number of pairs without counting selves
		$cij /= $cp - 1; # now divide by the total number of non-self domains in this protein
		$dt2dt2c{$acci}{$accj} += $cij*$numInstances if $cij; # add score to this running sum
	    }
	}
    }
    # last infered number is the dimensions of the matrix (the number of domains types that we observed)
    # should be at most the number of models in the Pfam database, but some models are never observed because we eliminate architectures, so the number will tend to be smaller than this total
    my $n = scalar keys %dt2c;
    return ($u, $n, \%dt2c, \%dt2dt2c); # done!  return all counts
}

sub getFirstOrderP {
    my ($dt2c, $n) = @_;
    # since we're using a uniform probability, this function is very boring.  Precalculate E = -log(P), P = 1/n, to copy everywhere
    my $E = General::log2($n);
    # the structures we want
    my %dt2E;
    # start navigating observed accessions in architectures, copy values
    foreach my $dt (keys %$dt2c) {
	$dt2E{$dt} = $E;
    }
    # done, return structure we want
    return (\%dt2E);
}

sub getSecondOrderPS {
    my ($dt2dt2c, $dt2E, $a, $ua, $n) = @_;
    # handy constant
    my $an2 = $a/($n*$n); # a divided by n squared, occurs often, should not repeat computation
    # structure that map pairs to scores (log likelihood of context vs null)
    my %dt2dt2S;
    while (my ($dti, $dt2c) = each %$dt2dt2c) {
	my $Ei = $dt2E->{$dti}; # energy of domain i
	while (my ($dtj, $cij) = each %$dt2c) {
	    # compute probability and score
	    my $Ej = $dt2E->{$dtj}; # energy of domain j
	    my $Pij = ($cij + $an2)/$ua; # probability of pair
	    my $Sij = General::log2($Pij)+$Ei+$Ej; # pair score (log likelihood of context hypothesis vs null), this is a shortcut using the Energy definition (-log(P))
	    
	    # add values to net
	    # note i and j are already sorted consistently i le j from their creation above
	    $dt2dt2S{$dti}{$dtj} = $Sij;
	}
    }
    # done, return structure we want
    return (\%dt2dt2S);
}
